using System;
using System.Collections.Generic;
using System.Configuration;
using System.Data.SqlClient;
using System.IO;
using System.Reflection;
using System.ServiceProcess;
using ITHit.WebDAV.Server.Logger;

namespace WebDAVServer.Class1Listener
{
    /// <summary>
    /// WebDAV engine host.
    /// </summary>
    internal class Program
    {
        /// <summary>
        /// Whether requests/responses shall be logged.
        /// </summary>
        private static readonly bool debugLoggingEnabled =
            ConfigurationManager.AppSettings["DebugLoggingEnabled"].Equals(
                "true",
                StringComparison.InvariantCultureIgnoreCase);

        /// <summary>
        /// Logger instance.
        /// </summary>
        private static readonly DefaultLoggerImpl logger = new DefaultLoggerImpl();

        /// <summary>
        /// Gets a value indicating whether the program is run as Windows service or standalone application.
        /// </summary>
        private static bool isServiceMode
        {
            get { return !Environment.UserInteractive; }
        }

        /// <summary>
        /// Entry point.
        /// </summary>
        /// <param name="args">Command line arguments.</param>
        public static void Main(string[] args)
        {
            var exePath = Path.GetDirectoryName(new Uri(Assembly.GetExecutingAssembly().CodeBase).LocalPath);
            logger.LogFile = Path.Combine(exePath, "WebDAVlog.txt");
            logger.IsDebugEnabled = debugLoggingEnabled;

            try
            {
                checkConfigErrors();

                if (isServiceMode)
                {
                    ServiceBase.Run(new Service(logger));
                }
                else
                {
                    (new Service(logger, true)).ThreadProc();
                }
            }
            catch (Exception ex)
            {
                if (isServiceMode)
                {
                    string errorMessage = "Could not start service\n     " + ex.Message;
                    logger.LogError(errorMessage, null);
                }
                else
                {
                    Console.WriteLine(ex.Message);
                    Console.WriteLine("\nPress any key...");
                    Console.ReadKey();
                }
            }
        }

        /// <summary>
        /// Checks configuration errors.
        /// </summary>
        private static void checkConfigErrors()
        {
            string connStr = ConfigurationManager.ConnectionStrings["WebDAV"].ConnectionString;
            if (string.IsNullOrEmpty(connStr))
            {
                throw new Exception("SqlConnectionString section is missing or invalid!");
            }

            try
            {
                using (SqlConnection conn = new SqlConnection(connStr))
                {
                    conn.Open();
                }
            }
            catch (Exception ex)
            {
                throw new Exception("SqlConnectionString configuration parameter contains invalid value.", ex);
            }

            string uriPrefix = ConfigurationManager.AppSettings["ListenerPrefix"];
            if (string.IsNullOrEmpty(uriPrefix))
            {
                throw new Exception("ListenerPrefix section is missing or invalid!");
            }

            string authType = ConfigurationManager.AppSettings["AuthenticationType"];
            if (authType == null)
            {
                throw new Exception("AuthenticationType configuration parameter is not specified.");
            }

            switch (authType.ToLower())
            {
                case "none":
                case "basic":
                case "digest":
                case "ntlm":
                case "kerberos":
                    break;
                default:
                    throw new Exception(
                        "Unsupported AuthenticationType configuration parameter value.\n" +
                        "Possible values are: (None | Basic | Digest | Ntlm | Kerberos)");
            }
        }
    }
}